use crate::vulkan::{VkSurface, LOG_TARGET};
use ash::prelude::VkResult;
use ash::vk;

pub struct VkSwapchain {
    pub(super) swapchain_loader: ash::khr::swapchain::Device,
    swapchain: Option<vk::SwapchainKHR>,

    pub(super) desired_image_count: u32,
    pub(super) surface_format: vk::SurfaceFormatKHR,
    pub(super) surface_resolution: vk::Extent2D,
    pub(super) present_mode: vk::PresentModeKHR,
    pub(super) pre_transform: vk::SurfaceTransformFlagsKHR,
}

impl VkSwapchain {
    pub(super) fn new(
        swapchain_loader: ash::khr::swapchain::Device,
        desired_image_count: u32,
        surface_format: vk::SurfaceFormatKHR,
        surface_resolution: vk::Extent2D,
        present_mode: vk::PresentModeKHR,
        pre_transform: vk::SurfaceTransformFlagsKHR,
    ) -> Self {
        Self {
            swapchain_loader,
            desired_image_count,
            surface_format,
            surface_resolution,
            present_mode,
            pre_transform,
            swapchain: None,
        }
    }

    pub(super) fn create_swapchain(&mut self, surface: &VkSurface) -> VkResult<()> {
        let mut swapchain_create_info = self.create_swapchain_info(surface);

        if let Some(old_swapchain) = self.swapchain {
            swapchain_create_info.old_swapchain = old_swapchain;
        }

        let swapchain = unsafe {
            self.swapchain_loader.create_swapchain(&swapchain_create_info, None)?
        };

        match self.swapchain {
            Some(_) => log::debug!(target: LOG_TARGET, "Swapchain created : {swapchain_create_info:#?}"),
            None => log::debug!(target: LOG_TARGET, "Swapchain updated : {swapchain_create_info:#?}")
        }

        self.swapchain = Some(swapchain);

        Ok(())
    }

    pub(super) fn update_resolution(&mut self, surface: &VkSurface, width: u32, height: u32) -> VkResult<()> {
        log::debug!(target: LOG_TARGET, "New resolution requested for swapchain {width}x{height}");
        self.surface_resolution = vk::Extent2D {
            width,
            height,
        };

        self.create_swapchain(surface)?;

        Ok(())
    }

    pub(super) fn get_swapchain_images(&self) -> anyhow::Result<Vec<vk::Image>> {
        let swapchain = self.swapchain
            .ok_or_else(|| anyhow::anyhow!("Can't get swapchain images : Swapchain is not set"))?;

        let images = unsafe { self.swapchain_loader.get_swapchain_images(swapchain)? };

        Ok(images)
    }

    fn create_swapchain_info(&self, surface: &VkSurface) -> vk::SwapchainCreateInfoKHR {
        vk::SwapchainCreateInfoKHR::default()
            .surface(surface.surface)
            .min_image_count(self.desired_image_count)
            .image_color_space(self.surface_format.color_space)
            .image_format(self.surface_format.format)
            .image_extent(self.surface_resolution)
            .image_usage(vk::ImageUsageFlags::COLOR_ATTACHMENT)
            .image_sharing_mode(vk::SharingMode::EXCLUSIVE)
            .pre_transform(self.pre_transform)
            .composite_alpha(vk::CompositeAlphaFlagsKHR::OPAQUE)
            .present_mode(self.present_mode)
            .clipped(true)
            .image_array_layers(1)
    }

    fn drop_current_swapchain(&mut self) {
        if let Some(swapchain) = self.swapchain {
            unsafe {
                self.swapchain_loader.destroy_swapchain(swapchain, None);
                self.swapchain = None;
            }
        }
    }
}

impl Drop for VkSwapchain {
    fn drop(&mut self) {
        self.drop_current_swapchain()
    }
}